from typing import List, Dict, Optional, Tuple
from .utils.solver_utils import SolverUtils
from .utils.dataclasses_utils import DataclassesUtils
from .dataclasses import Comparison, Criterion, DataValidator, Position, Intensity


class Inconsistency(Exception):
    def __init__(self, message, data=None):
        super().__init__(message)
        self.data = data


class Solver:

    def __init__(self, show_logs: Optional[bool] = False):
        self.name = 'UTA GMS Solver'
        self.show_logs = show_logs

    def __str__(self):
        return self.name

    def get_hasse_diagram_dict(
            self,
            performance_table_dict: Dict[str, Dict[str, float]],
            comparisons: List[Comparison],
            criteria: List[Criterion],
            positions: Optional[List[Position]] = [],
            intensities: Optional[List[Intensity]] = [],
    ) -> Dict[str, List[str]]:
        """
        Method for getting hasse diagram dict

        :param intensities:
        :param performance_table_dict:
        :param comparisons: List of Comparison objects
        :param criteria: List of Criterion objects
        :param positions: List of Position objects

        :return direct_relations:
        """
        DataValidator.validate_criteria(performance_table_dict, criteria)
        DataValidator.validate_performance_table(performance_table_dict)
        DataValidator.validate_positions(positions, performance_table_dict)

        refined_performance_table_dict: List[List[float]] = DataclassesUtils.refine_performance_table_dict(
            performance_table_dict=performance_table_dict
        )

        refined_comparisons: List[List[int]] = DataclassesUtils.refine_comparisons(
            performance_table_dict=performance_table_dict,
            comparisons=comparisons
        )

        refined_gains: List[bool] = DataclassesUtils.refine_gains(
            criterions=criteria
        )

        refined_linear_segments: List[int] = DataclassesUtils.refine_linear_segments(
            criterions=criteria
        )

        refined_worst_best_position: List[List[int]] = DataclassesUtils.refine_positions(
            positions=positions,
            performance_table_dict=performance_table_dict
        )

        refined_intensities: List[List[int]] = DataclassesUtils.refine_intensities(
            intensities=intensities,
            performance_table_dict=performance_table_dict
        )

        alternatives_id_list: List[str] = list(performance_table_dict.keys())

        necessary_preference = SolverUtils.get_necessary_relations(
            performance_table_list=refined_performance_table_dict,
            alternatives_id_list=alternatives_id_list,
            comparisons=refined_comparisons,
            criteria=refined_gains,
            worst_best_position=refined_worst_best_position,
            number_of_points=refined_linear_segments,
            comprehensive_intensities=refined_intensities,
            show_logs=self.show_logs
        )

        direct_relations: Dict[str, List[str]] = SolverUtils.calculate_direct_relations(necessary_preference)

        for alternatives_id in alternatives_id_list:
            if alternatives_id not in direct_relations.keys():
                direct_relations[alternatives_id] = []

        return direct_relations

    def get_representative_value_function_dict(
            self,
            performance_table_dict: Dict[str, Dict[str, float]],
            comparisons: List[Comparison],
            criteria: List[Criterion],
            positions: Optional[List[Position]] = [],
            intensities: Optional[List[Intensity]] = [],
            sampler_path: str = 'files/polyrun-1.1.0-jar-with-dependencies.jar',
            number_of_samples: str = '100',
            sampler_on: bool = True
    ) -> Tuple[Dict[str, float], Dict[str, List[Tuple[float, float]]], Dict[str, List[float]], Dict[str, Dict[str, float]], int, List[List[int]], Dict[str, List[str]], Dict[str, List[str]], str]:
        """
        Method for getting The Most Representative Value Function

        :param intensities:
        :param performance_table_dict:
        :param comparisons: List of Comparison objects
        :param criteria: List of Criterion objects
        :param positions: List of Position objects
        :param sampler_path:
        :param number_of_samples:
        :param sampler_on:

        :return:
        """
        DataValidator.validate_criteria(performance_table_dict, criteria)
        DataValidator.validate_performance_table(performance_table_dict)
        DataValidator.validate_positions(positions, performance_table_dict)
        DataValidator.validate_comparisons_criteria(comparisons, positions, criteria)

        refined_performance_table_dict: List[List[float]] = DataclassesUtils.refine_performance_table_dict(
            performance_table_dict=performance_table_dict
        )

        refined_comparisons: List[List[int]] = DataclassesUtils.refine_comparisons(
            performance_table_dict=performance_table_dict,
            comparisons=comparisons
        )

        refined_gains: List[bool] = DataclassesUtils.refine_gains(
            criterions=criteria
        )

        refined_linear_segments: List[int] = DataclassesUtils.refine_linear_segments(
            criterions=criteria
        )

        refined_worst_best_position: List[List[int]] = DataclassesUtils.refine_positions(
            positions=positions,
            performance_table_dict=performance_table_dict
        )

        refined_intensities: List[List[int]] = DataclassesUtils.refine_intensities(
            intensities=intensities,
            performance_table_dict=performance_table_dict
        )

        alternatives_id_list: List[str] = list(performance_table_dict.keys())

        problem, position_percentage, pairwise_percentage, number_of_samples_used, sampler_error = SolverUtils.calculate_the_most_representative_function(
            performance_table_list=refined_performance_table_dict,
            alternatives_id_list=alternatives_id_list,
            comparisons=refined_comparisons,
            criteria=refined_gains,
            worst_best_position=refined_worst_best_position,
            number_of_points=refined_linear_segments,
            comprehensive_intensities=refined_intensities,
            show_logs=self.show_logs,
            sampler_path=sampler_path,
            number_of_samples=number_of_samples,
            sampler_on=sampler_on
        )

        extreme_ranking = SolverUtils.calculate_extreme_ranking_analysis(
            performance_table_list=refined_performance_table_dict,
            comparisons=refined_comparisons,
            criteria=refined_gains,
            worst_best_position=refined_worst_best_position,
            number_of_points=refined_linear_segments,
            comprehensive_intensities=refined_intensities,
            show_logs=self.show_logs
        )

        refined_extreme_ranking: List[List[int]] = DataclassesUtils.refine_extreme_ranking(
            extreme_ranking=extreme_ranking,
            performance_table_dict=performance_table_dict
        )

        necessary, possible = SolverUtils.calculate_necessary_and_possible_relation_matrix(
            performance_table_list=refined_performance_table_dict,
            alternatives_id_list=alternatives_id_list,
            comparisons=refined_comparisons,
            criteria=refined_gains,
            worst_best_position=refined_worst_best_position,
            number_of_points=refined_linear_segments,
            comprehensive_intensities=refined_intensities,
            show_logs=self.show_logs
        )

        for variable in problem.variables():
            if variable.name == 'epsilon':
                if variable.varValue <= 0.0:
                    resolved_inconsistencies = SolverUtils.resolve_incosistency(
                        performance_table_list=refined_performance_table_dict,
                        comparisons=refined_comparisons,
                        criteria=refined_gains,
                        worst_best_position=refined_worst_best_position,
                        number_of_points=refined_linear_segments,
                        comprehensive_intensities=refined_intensities,
                        subsets_to_remove=[],
                        show_logs=self.show_logs
                    )

                    refined_resolved_inconsistencies = DataclassesUtils.refine_resolved_inconsistencies(
                        resolved_inconsistencies=resolved_inconsistencies,
                        performance_table_dict=performance_table_dict
                    )

                    raise Inconsistency("Found inconsistencies", refined_resolved_inconsistencies)
                break

        variables_and_values_dict: Dict[str, float] = {variable.name: variable.varValue for variable in problem.variables()}

        criterion_functions: Dict[str, List[Tuple[float, float]]] = SolverUtils.get_criterion_functions(
            variables_and_values_dict=variables_and_values_dict,
            criteria=criteria,
        )

        alternatives_and_utilities_dict: Dict[str, float] = SolverUtils.get_alternatives_and_utilities_dict(
            variables_and_values_dict=variables_and_values_dict,
            performance_table_list=refined_performance_table_dict,
            alternatives_id_list=alternatives_id_list,
        )

        return alternatives_and_utilities_dict, criterion_functions, position_percentage, pairwise_percentage, number_of_samples_used, refined_extreme_ranking, necessary, possible, sampler_error



import pytest
from src.utagmsengine.utils.dataclasses_utils import DataclassesUtils
from src.utagmsengine.dataclasses import Comparison, Criterion


@pytest.fixture()
def refined_performance_table_dict_dummy():
    return [[26.0, 40.0, 44.0],  # 0 - A
            [2.0, 2.0, 68.0],  # 1 - B
            [18.0, 17.0, 14.0],  # 2 - C
            [35.0, 62.0, 25.0],  # 3 - D
            [7.0, 55.0, 12.0],  # 4 - E
            [25.0, 30.0, 12.0],  # 5 - F
            [9.0, 62.0, 88.0],  # 6 - G
            [0.0, 24.0, 73.0],  # 7 - H
            [6.0, 15.0, 100.0],  # 8 - I
            [16.0, 9.0, 0.0],  # 9 - J
            [26.0, 17.0, 17.0],  # 10 - K
            [62.0, 43.0, 0.0]]  # 11 - L


@pytest.fixture()
def performance_table_dict_dummy():
    return {
        'A': {'g1': 26.0, 'g2': 40.0, 'g3': 44.0},
        'B': {'g1': 2.0, 'g2': 2.0, 'g3': 68.0},
        'C': {'g1': 18.0, 'g2': 17.0, 'g3': 14.0},
        'D': {'g1': 35.0, 'g2': 62.0, 'g3': 25.0},
        'E': {'g1': 7.0, 'g2': 55.0, 'g3': 12.0},
        'F': {'g1': 25.0, 'g2': 30.0, 'g3': 12.0},
        'G': {'g1': 9.0, 'g2': 62.0, 'g3': 88.0},
        'H': {'g1': 0.0, 'g2': 24.0, 'g3': 73.0},
        'I': {'g1': 6.0, 'g2': 15.0, 'g3': 100.0},
        'J': {'g1': 16.0, 'g2': 9.0, 'g3': 0.0},
        'K': {'g1': 26.0, 'g2': 17.0, 'g3': 17.0},
        'L': {'g1': 62.0, 'g2': 43.0, 'g3': 0.0}
    }


@pytest.fixture()
def refined_comparisons_dummy():
    return [
        [6, 5, [], '>'],
        [5, 4, [], '>'],
        [3, 6, [], '=']
    ]


@pytest.fixture()
def comparisons_dummy():
    return [
        Comparison(alternative_1='G', alternative_2='F'),
        Comparison(alternative_1='F', alternative_2='E'),
        Comparison(alternative_1='D', alternative_2='G', sign='=')
    ]


@pytest.fixture()
def criterions_dummy():
    return [Criterion(criterion_id='1', gain=True), Criterion(criterion_id='1', gain=True), Criterion(criterion_id='1', gain=True)]


def test_refine_performance_table_dict(
        performance_table_dict_dummy,
        refined_performance_table_dict_dummy
):
    refined_performance_table_dict = DataclassesUtils.refine_performance_table_dict(
        performance_table_dict=performance_table_dict_dummy
    )

    assert refined_performance_table_dict == refined_performance_table_dict_dummy


def test_refine_comparisons(
        performance_table_dict_dummy,
        comparisons_dummy,
        refined_comparisons_dummy
):
    refined_comparisons = DataclassesUtils.refine_comparisons(
        performance_table_dict=performance_table_dict_dummy,
        comparisons=comparisons_dummy
    )

    assert refined_comparisons == refined_comparisons_dummy






import pytest
from src.utagmsengine.dataclasses import Comparison, Criterion, DataValidator, Position


@pytest.fixture()
def performance_table_list_dummy():
    return {
        'A': {'g1': 26.0, 'g2': 40.0, 'g4': 44.0},
        'B': {'a1': 2.0, 'g2': 2.0, 'g4': 68.0},
        'C': {'g1': 18.0, 'g2': 17.0, 'g4': 14.0},
        'D': {'g1': 35.0, 'g2': 62.0, 'g4': 25.0},
        'E': {'g1': 7.0, 'g2': 55.0, 'g4': 12.0},
        'F': {'g1': 25.0, 'g2': 30.0, 'g4': 12.0},
        'G': {'g1': 9.0, 'g2': 62.0, 'g4': 88.0},
        'H': {'g1': 0.0, 'g2': 24.0, 'g4': 73.0},
        'I': {'g1': 6.0, 'g2': 15.0, 'g4': 100.0},
        'J': {'g1': 16.0, 'g2': 9.0, 'g4': 0.0},
        'K': {'g1': 26.0, 'g2': 17.0, 'g4': 17.0},
        'L': {'g1': 62.0, 'g2': 43.0, 'g4': 0.0}
    }


@pytest.fixture()
def comparisons_list_dummy():
    return [
        {'alternative_1': 'G', 'alternative_2': 'F'},
        {'alternative_1': 'F', 'alternative_2': 'E'},
        {'alternative_1': 'D', 'alternative_2': 'G', 'sign': '='}
    ]


@pytest.fixture()
def criterion_list_dummy():
    return [
        {'criterion_id': 'g1', 'gain': True, 'number_of_linear_segments': 0},
        {'criterion_id': 'g2', 'gain': True, 'number_of_linear_segments': 0},
        {'criterion_id': 'g3', 'gain': True, 'number_of_linear_segments': 0},
    ]


@pytest.fixture()
def comparisons_dummy():
    return [
        Comparison(alternative_1='G', alternative_2='F', criteria=[], sign='>'),
        Comparison(alternative_1='F', alternative_2='E', criteria=[], sign='>'),
        Comparison(alternative_1='D', alternative_2='G', criteria=[], sign='=')
    ]


@pytest.fixture()
def criterions_dummy():
    return [Criterion(criterion_id='g1', gain=True, number_of_linear_segments=0), Criterion(criterion_id='g2', gain=True, number_of_linear_segments=0), Criterion(criterion_id='g3', gain=True, number_of_linear_segments=0)]


@pytest.fixture()
def positions_list_dummy():
    return [Position(alternative_id='M', worst_position=1, best_position=3)]


def test_comparisons(
        comparisons_list_dummy,
        comparisons_dummy
):
    comparisons = [Comparison(**data) for data in comparisons_list_dummy]

    assert comparisons == comparisons_dummy


def test_criterions(
        criterion_list_dummy,
        criterions_dummy
):
    criterions = [Criterion(**data) for data in criterion_list_dummy]

    assert criterions == criterions_dummy


def test_comparison_validation():
    with pytest.raises(ValueError, match="alternative_1 and alternative_2 options must be different."):
        Comparison(alternative_1='A', alternative_2='A')


def test_criterion_linear_segments_validation():
    with pytest.raises(ValueError, match="Number of linear segments can't be negative."):
        Criterion(criterion_id='g1', gain=True, number_of_linear_segments=-1)


def test_data_validator_validate_criteria(
        performance_table_list_dummy,
        criterions_dummy
):
    with pytest.raises(ValueError, match="Criterion IDs in the list and the data dictionary do not match."):
        DataValidator.validate_criteria(performance_table_list_dummy, criterions_dummy)


def test_data_validator_validate_performance_table(
        performance_table_list_dummy
):
    with pytest.raises(ValueError, match="Keys inside the inner dictionaries are not consistent."):
        DataValidator.validate_performance_table(performance_table_list_dummy)


def test_data_validator_validate_positions(
        positions_list_dummy,
        performance_table_list_dummy
):
    with pytest.raises(ValueError, match="Alternative IDs in the Position list and the data dictionary do not match."):
        DataValidator.validate_positions(positions_list_dummy, performance_table_list_dummy)

    with pytest.raises(ValueError, match="worst_position can't be negative."):
        Position(alternative_id='A', worst_position=-1, best_position=2)

    with pytest.raises(ValueError, match="best_position can't be negative."):
        Position(alternative_id='A', worst_position=1, best_position=-2)



import pytest
from src.utagmsengine.solver import Solver, Inconsistency
from src.utagmsengine.dataclasses import Comparison, Criterion, Position, Intensity


@pytest.fixture()
def performance_table_dict_dummy():
    return {
        'A': {'g1': 26.0, 'g2': 40.0, 'g3': 44.0},
        'B': {'g1': 2.0, 'g2': 2.0, 'g3': 68.0},
        'C': {'g1': 18.0, 'g2': 17.0, 'g3': 14.0},
        'D': {'g1': 35.0, 'g2': 62.0, 'g3': 25.0},
        'E': {'g1': -7.0, 'g2': 55.0, 'g3': 12.0},
        'F': {'g1': 25.0, 'g2': 30.0, 'g3': 12.0},
        'G': {'g1': 9.0, 'g2': 62.0, 'g3': 88.0},
        'H': {'g1': 0.0, 'g2': 24.0, 'g3': 73.0},
        'I': {'g1': 6.0, 'g2': 15.0, 'g3': 100.0},
        'J': {'g1': 16.0, 'g2': -9.0, 'g3': 0.0},
        'K': {'g1': 26.0, 'g2': 17.0, 'g3': 17.0},
        'L': {'g1': 62.0, 'g2': 43.0, 'g3': 0.0}
    }


@pytest.fixture()
def comparison_dummy():
    return [
        Comparison(alternative_1='G', alternative_2='F', sign='>'),
        Comparison(alternative_1='F', alternative_2='E', sign='>'),
        Comparison(alternative_1='D', alternative_2='G', sign='='),
    ]


@pytest.fixture()
def criterions_dummy():
    return [Criterion(criterion_id='g1', gain=True, number_of_linear_segments=0), Criterion(criterion_id='g2', gain=True, number_of_linear_segments=0), Criterion(criterion_id='g3', gain=True, number_of_linear_segments=0)]


@pytest.fixture()
def predefined_criterions_dummy():
    return [Criterion(criterion_id='g1', gain=True, number_of_linear_segments=4), Criterion(criterion_id='g2', gain=True, number_of_linear_segments=4), Criterion(criterion_id='g3', gain=True, number_of_linear_segments=4)]


@pytest.fixture()
def positions_dummy():
    return [Position(alternative_id='A', worst_position=12, best_position=2)]


@pytest.fixture()
def intensities_dummy():
    return [Intensity(alternative_id_1='H', alternative_id_2='G', alternative_id_3='B', alternative_id_4='D', criteria=['g1', 'g2'])]


@pytest.fixture()
def resolved_inconsistencies_dummy():
    return [[[], [], [Position(alternative_id='A', worst_position=12, best_position=9, criteria=[])], []], [[Comparison(alternative_1='G', alternative_2='F', criteria=[], sign='>')], [], [], []], [[], [Comparison(alternative_1='D', alternative_2='G', criteria=[], sign='=')], [], []]]


@pytest.fixture()
def hasse_diagram_dict_dummy():
    return {'A': ['F', 'K'], 'C': ['J'], 'D': ['G'], 'F': ['E', 'J'], 'G': ['B', 'D', 'F', 'H', 'K'], 'I': ['B'], 'K': ['C'], 'L': ['J'], 'B': [], 'E': [], 'H': [], 'J': []}


@pytest.fixture()
def representative_value_function_dict_dummy():
    return {'J': 0.0738, 'E': 0.1955, 'L': 0.2244, 'C': 0.3235, 'F': 0.382, 'B': 0.4467, 'H': 0.4523, 'K': 0.4741, 'A': 0.6248, 'D': 0.6248, 'G': 0.6248, 'I': 0.7754}


@pytest.fixture()
def criterion_functions_dummy():
    return {'g1': [(-7.0, 0.0), (0.0, 0.0), (2.0, 0.0), (6.0, 0.0), (9.0, 0.0), (10.25, 0.0), (16.0, 0.0737571), (18.0, 0.0994118), (25.0, 0.189203), (26.0, 0.20203), (27.5, 0.221271), (35.0, 0.221271), (44.75, 0.221271), (62.0, 0.221271)], 'g2': [(-9.0, 0.0), (2.0, 0.0), (8.75, 0.0), (15.0, 0.0), (17.0, 0.0), (24.0, 0.0), (26.5, 0.0), (30.0, 0.000661446), (40.0, 0.00255129), (43.0, 0.00311825), (44.25, 0.00335448), (55.0, 0.00335448), (62.0, 0.00335448)], 'g3': [(0.0, 0.0), (12.0, 0.192056), (14.0, 0.224065), (17.0, 0.272079), (25.0, 0.400116), (44.0, 0.42016), (50.0, 0.42649), (68.0, 0.446706), (73.0, 0.452322), (75.0, 0.454568), (88.0, 0.621387), (100.0, 0.775374)]}


@pytest.fixture()
def predefined_hasse_diagram_dict_dummy():
    return {'A': ['F', 'K'], 'C': ['J'], 'D': ['G'], 'F': ['E', 'J'], 'G': ['B', 'D', 'F', 'H', 'K'], 'I': ['B'], 'K': ['C'], 'L': ['J'], 'B': [], 'E': [], 'H': [], 'J': []}


@pytest.fixture()
def extreme_ranking_dummy():
    return {'A': ((11, 2), (7, 2)), 'B': ((12, 4), (12, 1)), 'C': ((12, 5), (11, 3)), 'D': ((7, 2), (4, 1)), 'E': ((12, 5), (12, 5)), 'F': ((11, 4), (10, 3)), 'G': ((7, 2), (4, 1)), 'H': ((12, 3), (12, 1)), 'I': ((11, 1), (10, 1)), 'J': ((12, 8), (12, 4)), 'K': ((12, 4), (10, 3)), 'L': ((12, 1), (11, 1))}


def test_get_hasse_diagram_dict(
        performance_table_dict_dummy,
        comparison_dummy,
        criterions_dummy,
        positions_dummy,
        intensities_dummy,
        hasse_diagram_dict_dummy
):
    solver = Solver(show_logs=True)

    hasse_diagram_list = solver.get_hasse_diagram_dict(
        performance_table_dict_dummy,
        comparison_dummy,
        criterions_dummy,
        positions_dummy,
        intensities_dummy
    )

    assert hasse_diagram_list == hasse_diagram_dict_dummy


def test_get_representative_value_function_dict(
        performance_table_dict_dummy,
        comparison_dummy,
        predefined_criterions_dummy,
        criterions_dummy,
        positions_dummy,
        intensities_dummy,
        representative_value_function_dict_dummy,
        criterion_functions_dummy,
        resolved_inconsistencies_dummy,
        extreme_ranking_dummy
):
    solver = Solver(show_logs=True)

    try:
        representative_value_function_dict, criterion_functions, position_percentage, pairwise_percentage, number_of_rejected, extreme_ranking, necessary, possible, sampler_error = (
            solver.get_representative_value_function_dict(
                performance_table_dict_dummy,
                comparison_dummy,
                predefined_criterions_dummy,
                positions_dummy,
                intensities_dummy,
                'files/polyrun-1.1.0-jar-with-dependencies.jar',
                '10'
            )
        )

        assert representative_value_function_dict == representative_value_function_dict_dummy
        assert criterion_functions == criterion_functions_dummy
        assert extreme_ranking == extreme_ranking_dummy

    except Inconsistency as e:
        resolved_inconsistencies = e.data

        assert resolved_inconsistencies == resolved_inconsistencies_dummy


def test_predefined_get_hasse_diagram_dict(
        performance_table_dict_dummy,
        comparison_dummy,
        predefined_criterions_dummy,
        positions_dummy,
        predefined_hasse_diagram_dict_dummy
):
    solver = Solver(show_logs=True)

    hasse_diagram_list = solver.get_hasse_diagram_dict(
        performance_table_dict_dummy,
        comparison_dummy,
        predefined_criterions_dummy,
        positions_dummy
    )

    assert hasse_diagram_list == predefined_hasse_diagram_dict_dummy



from xmcda.XMCDA import XMCDA
from src.utagmsengine.utils.parser_utils import ParserUtils
import _io


def test_load_file():
    with open('tests/files/performance_table.xml', 'rb') as xml_file:
        xml_file_text = _io.TextIOWrapper(xml_file, encoding='utf-8')
        xmcda: XMCDA = ParserUtils.load_xmcda(xml_file_text)

        assert xmcda.alternatives[0].id == 'A'
        assert xmcda.alternatives[5].id == 'F'
        assert xmcda.criteria[0].id == 'g1'



from typing import List, Dict, Tuple
from ..dataclasses import Comparison, Position, Intensity


class DataclassesUtils:

    @staticmethod
    def refine_performance_table_dict(
            performance_table_dict: Dict[str, Dict[str, float]]
    ) -> List[List[float]]:
        """
        Convert a dictionary of performance table values into a 2D list of floats.

        :param performance_table_dict:
        :return output_list:
        """
        output_list = []
        for key in performance_table_dict:
            inner_list = list(performance_table_dict[key].values())
            output_list.append(inner_list)

        return output_list

    @staticmethod
    def refine_comparisons(
            performance_table_dict: Dict[str, Dict[str, float]],
            comparisons
    ) -> List[List[int]]:
        """
        Convert a list of Comparison into a list of indices corresponding to alternatives.

        :param performance_table_dict:
        :param comparisons:
        :return output:
        """
        output = []
        keys: List[str] = list(performance_table_dict.keys())

        for comparison in comparisons:
            alternative_1_index: int = keys.index(comparison.alternative_1)
            alternative_2_index: int = keys.index(comparison.alternative_2)
            sign: str = comparison.sign

            first: Dict[str, float] = next(iter(performance_table_dict.values()))  # Get the first dictionary value
            criteria_index: List[int] = []
            for criteria in comparison.criteria:
                position = list(first.keys()).index(criteria)
                criteria_index.append(position)

            output.append([alternative_1_index, alternative_2_index, criteria_index, sign])

        return output

    @staticmethod
    def refine_gains(
            criterions
    ) -> List[bool]:
        """
        Extract gains/costs from a list of Criterion objects.

        :param criterions:

        :return output:
        """
        output = []

        for criterion in criterions:
            output.append(criterion.gain)

        return output

    @staticmethod
    def refine_linear_segments(
            criterions
    ) -> List[int]:
        """
        Extract number of linear segments from a list of Criterion objects.

        :param criterions:

        :return output:
        """
        output = []

        for criterion in criterions:
            if criterion.number_of_linear_segments == 0:
                linear_segments = 0
            else:
                linear_segments = criterion.number_of_linear_segments + 1
            output.append(linear_segments)

        return output

    @staticmethod
    def refine_positions(
            positions,
            performance_table_dict
    ) -> List[List[int]]:
        """
        Refined list[Positions] to [[alternative_ID, worst_position, best_position], ...] format

        :param positions:
        :param performance_table_dict:

        :return output:
        """
        output = []
        tmp = {}

        for i, key in enumerate(performance_table_dict.keys()):
            tmp[key] = i

        for position in positions:
            first = next(iter(performance_table_dict.values()))  # Get the first dictionary value
            criteria_index = []
            for criteria in position.criteria:
                index = list(first.keys()).index(criteria)
                criteria_index.append(index)

            output.append([tmp[position.alternative_id], position.worst_position, position.best_position, criteria_index])

        return output

    @staticmethod
    def refine_intensities(
            intensities,
            performance_table_dict
    ) -> List[List[int]]:
        """
        Refine into [[1,[],3,[],4,[],5,[],'>']], meaning 1-3 > 4-5

        :param intensities:
        :param performance_table_dict:

        :return output:
        """
        output = []
        tmp = {}

        for i, key in enumerate(performance_table_dict.keys()):
            tmp[key] = i

        for intensity in intensities:
            first = next(iter(performance_table_dict.values()))  # Get the first dictionary value
            criteria_index = []
            for criteria in intensity.criteria:
                index = list(first.keys()).index(criteria)
                criteria_index.append(index)

            output.append(
                [
                    tmp[intensity.alternative_id_1],
                    criteria_index,
                    tmp[intensity.alternative_id_2],
                    criteria_index,
                    tmp[intensity.alternative_id_3],
                    criteria_index,
                    tmp[intensity.alternative_id_4],
                    criteria_index,
                    intensity.sign
                ]
            )

        return output

    @staticmethod
    def refine_resolved_inconsistencies(
            resolved_inconsistencies,
            performance_table_dict,
    ) -> List[List[int]]:
        """
        Refine from [[1,[],3,[],4,[],5,[],'>']], meaning 1-3 > 4-5 into dataclasses

        :param resolved_inconsistencies:
        :param performance_table_dict:

        :return output:
        """
        output = []
        alt_idx = {}
        crit_idx = {}

        for i, key in enumerate(performance_table_dict.keys()):
            alt_idx[i] = key

        for i, key in enumerate(performance_table_dict[alt_idx[0]]):
            crit_idx[i] = key

        for inconsistencies in resolved_inconsistencies:
            comparisons = []
            positions = []
            intensities = []

            for comparison in inconsistencies[0]:
                comparisons.append(
                    Comparison(alternative_1=alt_idx[comparison[0]], alternative_2=alt_idx[comparison[1]], criteria=[crit_idx[idx] for idx in comparison[2]], sign=comparison[3])
                )

            for position in inconsistencies[1]:
                positions.append(
                    Position(alternative_id=alt_idx[position[0]], worst_position=position[1], best_position=position[2], criteria=[crit_idx[idx] for idx in position[3]])
                )

            for intensity in inconsistencies[2]:
                intensities.append(
                    Intensity(
                        alternative_id_1=alt_idx[intensity[0]],
                        alternative_id_2=alt_idx[intensity[2]],
                        alternative_id_3=alt_idx[intensity[4]],
                        alternative_id_4=alt_idx[intensity[6]],
                        criteria=[crit_idx[idx] for idx in intensity[1]],
                        sign=intensity[8]
                    )
                )

            output.append(
                [
                    comparisons,
                    positions,
                    intensities
                 ]
            )

        return output[:-1]

    @staticmethod
    def refine_extreme_ranking(
            extreme_ranking,
            performance_table_dict
    ) -> List[List[int]]:
        """
        Refine extreme_ranking

        :param extreme_ranking:
        :param performance_table_dict:

        :return output:
        """
        output: Dict[str, Tuple[Tuple[int]]] = {}
        tmp = {}

        for i, key in enumerate(performance_table_dict.keys()):
            tmp[i] = key

        for extreme_rank in extreme_ranking:
            pessimistic = (extreme_rank[1], extreme_rank[3])
            optimistic = (extreme_rank[2], extreme_rank[4])
            output[tmp[extreme_rank[0]]] = (pessimistic, optimistic)
        return output
